#include <stdio.h>
#include <ctype.h>
#include <fcntl.h>
#include <getopt.h>
#include <linux/types.h>
#include <linux/kd.h>
#include <linux/keyboard.h>
#include "ksyms.h"

#define VERSION "0.81"
static int fd;

int get_bind(u_char index, u_char table) {
	struct kbentry ke;

	ke.kb_index = index;
	ke.kb_table = table;
	if (ioctl(fd, KDGKBENT, (unsigned long)&ke)) {
		fprintf(stderr, "KDGKBENT at index %d in table %d", index, table);
		perror("");
		exit(1);
	}
	return ke.kb_value;
}

void print_keysym(int code, char numeric) {
	int t;
	int v;
	char *p;

	t = KTYP(code);
	v = KVAL(code);
	if (
		!numeric &&
		t < syms_size &&
		v < syms[t].size &&
		(p = syms[t].table[v])[0]
	)
		printf(" %-16s", p);
	else
		printf(" 0x%04x          ", code);
}

/* there's a race in here */
char valid_type(int t) {
	struct kbentry ke;
	int old_value;
	char status;

	ke.kb_index = 0;
	ke.kb_table = 0;
	if (ioctl(fd, KDGKBENT, (unsigned long)&ke))
		return 0;
	old_value = ke.kb_value;
	ke.kb_value = K(t, 0);
	status = ioctl(fd, KDSKBENT, (unsigned long)&ke) == 0;
	ke.kb_value = old_value;
	ioctl(fd, KDSKBENT, (unsigned long)&ke);
	return status;
}

/* there's a race in here */
u_char maximum_val(int t) {
	struct kbentry ke;
	int old_value;
	char status;
	int i;

	ke.kb_index = 0;
	ke.kb_table = 0;
	if (ioctl(fd, KDGKBENT, (unsigned long)&ke))
		return 0;
	old_value = ke.kb_value;

	for (i = 0; i < 256; i++) {
		ke.kb_value = K(t, i);
		if (ioctl(fd, KDSKBENT, (unsigned long)&ke))
			break;
	}

	ke.kb_value = old_value;
	ioctl(fd, KDSKBENT, (unsigned long)&ke);

	return i - 1;
}


void show_short_info(void) {
	int i;

	printf("keycode range supported by kernel:           0 - %d\n",
		NR_KEYS - 1);
	printf("number of actions bindable to a key:         %d\n",
		NR_KEYMAPS);
	printf("ranges of action codes supported by kernel:\n");
	for (i = 0; valid_type(i); i++)
		printf("	0x%04x - 0x%04x\n",
			K(i, 0), K(i, maximum_val(i)));
	printf("number of function keys supported by kernel: %d\n", NR_FUNC);
	printf("total space available for function strings:  %d bytes\n",
		FUNC_BUFSIZE);
}

void dump_symbols(void) {
	int t;
	int v;
	char *p;

	printf("Symbols recognized by dumpkeys:\n(numeric value, symbol)\n\n");
	for (t = 0; t < syms_size; t++)
		for (v = 0; v < syms[t].size; v++)
			if ((p = syms[t].table[v])[0])
				printf("0x%04x\t%s\n", K(t, v), p);
	printf("\
Recognized modifier names and their column numbers:

shift	%d
altgr	%d
control	%d
alt	%d
", (1 << KG_SHIFT), (1 << KG_ALTGR), (1 << KG_CTRL), (1 << KG_ALT));
}

void print_mod(int x) {
	if (x & (1 << KG_SHIFT))
		printf("shift   ");
	if (x & (1 << KG_ALTGR))
		printf("altgr   ");
	if (x & (1 << KG_CTRL))
		printf("control ");
	if (x & (1 << KG_ALT))
		printf("alt     ");
}

void print_bind(int buf[], int i, int j, char numeric) {
	printf("\t");
	print_mod(j);
	printf("keycode %3d =", i);
	print_keysym(buf[j], numeric);
	printf("\n");
}

void dump_keys(char full_table, char numeric) {
	int i;
	int j;
	int k;
	int buf[NR_KEYMAPS];
	int vmax;
	int vlast;
	int vcount;
	int count;

	for (i = 0; i < NR_KEYS; i++) {
		for (j = 0; j < NR_KEYMAPS; j++)
			buf[j] = get_bind(i, j);

		for (j = NR_KEYMAPS - 1; j >= 0; j--)
			if (buf[j] != K_HOLE)
				break;
		vmax = j;
		vlast = j;
		vcount = 0;
		for (j; j > 0; j--) {
			if (buf[j] != K_HOLE)
				vcount++;
			else if (vmax - j + 1 > 2 * vcount)
				vlast = j - 1;
		}

		count = 0;
		for (j = 1; j < NR_KEYMAPS; j++)
			if (buf[0] != buf[j])
				count++;

		if (full_table)
			vlast = NR_KEYMAPS - 1;

		printf("keycode %3d =", i);
		if (
			!full_table &&
			((buf[0] >= 'a' && buf[0] <= 'z') ||
			(buf[0] >= 'A' && buf[0] <= 'Z')) &&
			buf[1] == (buf[0] ^ 32) &&
			buf[2] == buf[0] &&
			buf[3] == (buf[0] ^ 32) &&
			buf[4] == (buf[0] & ~96) &&
			buf[5] == (buf[0] & ~96) &&
			buf[6] == (buf[0] & ~96) &&
			buf[7] == (buf[0] & ~96) &&
			KTYP(buf[8]) == KT_META && KVAL(buf[8]) == buf[0] &&
			KTYP(buf[9]) == KT_META && KVAL(buf[9]) == (buf[0] ^ 32) &&
			KTYP(buf[10]) == KT_META && KVAL(buf[10]) == buf[0] &&
			KTYP(buf[11]) == KT_META && KVAL(buf[11]) == (buf[0] ^ 32) &&
			KTYP(buf[12]) == KT_META && KVAL(buf[12]) == (buf[0] & ~96) &&
			KTYP(buf[13]) == KT_META && KVAL(buf[13]) == (buf[0] & ~96) &&
			KTYP(buf[14]) == KT_META && KVAL(buf[14]) == (buf[0] & ~96) &&
			KTYP(buf[15]) == KT_META && KVAL(buf[15]) == (buf[0] & ~96)
		) {
			print_keysym(buf[0], numeric);
			printf("\n");
		} else if (!full_table && count < NR_KEYMAPS / 2) {
			if (buf[0] != K_HOLE)
				print_keysym(buf[0], numeric);
			printf("\n");
			for (j = 1; j < NR_KEYMAPS; j++)
				if (buf[0] != buf[j])
					print_bind(buf, i, j, numeric);
		} else {
			for (k = 0; k <= vlast; k++)
				print_keysym(buf[k], numeric);
			printf("\n");
			for (; k <= vmax; k++)
				if (buf[k] != K_HOLE)
					print_bind(buf, i, k, numeric);
		}
	}
}

void dump_funcs(void) {
	int i;
	struct kbsentry fbuf;
	char *p;

	for (i = 0; i < NR_FUNC; i++) {
		fbuf.kb_func = i;
		if (ioctl(fd, KDGKBSENT, (unsigned long)&fbuf)) {
			fprintf(stderr, "KDGKBSENT at index %d", i);
			perror("");
			exit(1);
		}
		printf("string %s = \"", syms[KT_FN].table[i]);
		for (p = fbuf.kb_string; *p; p++)
			if (isgraph(*p))
				putchar(*p);
			else
				printf("\\%03o", *p);
		printf("\"\n");
	}
}

void usage(void) {
	fprintf(stderr, "\
dumpkeys version " VERSION "\

usage: dumpkeys [options...]

valid options are:

	-h --help	display this help text
	-i --short-info	display information about keyboard driver
	-l --long-info	display above and symbols known to loadkeys
	-n --numeric	display keytable in hexadecimal notation
	-f --full-table	don't use short-hand notations
	   --funcs-only	display only the function key strings
	   --keys-only	display only key bindings
	-c --charset={iso-8859-1,iso-8859-2,iso-8859-3,iso-8859-4}
			interpret character action codes to be from the
			specified character set
");
	exit(1);
}

main (int argc, char *argv[]) {
	const char *short_opts = "hilsnfc:";
	const struct option long_opts[] = {
		{ "help",	no_argument,		NULL, 'h' },
		{ "short-info",	no_argument,		NULL, 'i' },
		{ "long-info",	no_argument,		NULL, 'l' },
		{ "numeric",	no_argument,		NULL, 'n' },
		{ "full-table",	no_argument,		NULL, 'f' },
		{ "funcs-only",	no_argument,		NULL, 't' },
		{ "keys-only",	no_argument,		NULL, 'k' },
		{ "charset",	required_argument,	NULL, 'c' },
		{ NULL,	0, NULL, 0 }
	};
	int c;
	char long_info = 0;
	char short_info = 0;
	char numeric = 0;
	char full_table = 0;
	char funcs_only = 0;
	char keys_only = 0;
	char charset = 1;

	while ((c = getopt_long(argc, argv,
		short_opts, long_opts, NULL)) != -1) {
		switch (c) {
			case 'i':
				short_info = 1;
				break;
			case 's':
			case 'l':
				long_info = 1;
				break;
			case 'n':
				numeric = 1;
				break;
			case 'f':
				full_table = 1;
				break;
			case 't':
				funcs_only = 1;
				break;
			case 'k':
				keys_only = 1;
				break;
			case 'c':
				if (strlen(optarg) != 10 || /* iso-8859-x */
					strncmp(optarg, "iso-8859-", 9) ||
					optarg[9] < '1' || optarg[9] > '4')
					usage();
				charset = optarg[9] - '0';
				break;
			case 'h':
			case '?':
				usage();
		}
	}

	if (optind < argc)
		usage();

	if (charset > 1)
		syms_iso8859(charset);
	if ((fd = open("/dev/console", O_RDONLY)) < 0) {
		perror("when opening /dev/console");
		exit(1);
	}

	if (short_info || long_info) {
		show_short_info();
		if (long_info)
			dump_symbols();
		exit(0);
	}

	if (!funcs_only)
		dump_keys(full_table, numeric);
	if (!keys_only)
		dump_funcs();

	exit(0);
}
