#include <stdio.h>
#include <unistd.h>
#include <getopt.h>
#include <signal.h>
#include <fcntl.h>
#include <termios.h>
#include <linux/kd.h>

#define VERSION "0.81"

int tmp;	/* for debuggin */

#ifndef OLDKERNEL
#define OPTION "medium raw"
#else
#define OPTION "raw"

#include <linux/keyboard.h>

#define E0_BASE 96

unsigned long kbd_prev_dead_keys, kbd_dead_keys;

void translate(unsigned char scancode)
{
	char break_flag;
	static struct { char s:1; } key_down[NR_KEYS];
     	static unsigned char e0_keys[] = {
		0x1c,	/* keypad enter */
		0x1d,	/* right control */
		0x35,	/* keypad slash */
		0x37,	/* print screen */
		0x38,	/* right alt */
		0x46,	/* break (control-pause) */
		0x47,	/* editpad home */
		0x48,	/* editpad up */
		0x49,	/* editpad pgup */
		0x4b,	/* editpad left */
		0x4d,	/* editpad right */
		0x4f,	/* editpad end */
		0x50,	/* editpad dn */
		0x51,	/* editpad pgdn */
		0x52,	/* editpad ins */
		0x53,	/* editpad del */
#ifdef LK450
                0x3d,   /* f13 */
                0x3e,   /* f14 */
                0x3f,   /* help */
                0x40,   /* do */
                0x41,   /* f17 */
                0x4e    /* keypad minus/plus */
#endif
	};

	if (scancode == 0xe0) {
		set_kbd_dead(KGD_E0);
		return;
	}
	if (scancode == 0xe1) {
		set_kbd_dead(KGD_E1);
		return;
	}
	/*
	 *  The keyboard maintains its own internal caps lock and num lock
	 *  statuses. In caps lock mode E0 AA precedes make code and E0 2A
	 *  follows break code. In num lock mode, E0 2A precedes make
	 *  code and E0 AA follows break code. We do our own book-keeping,
	 *  so we will just ignore these.
	 */
	if (kbd_dead(KGD_E0) && (scancode == 0x2a || scancode == 0xaa ||
				 scancode == 0x36 || scancode == 0xb6))
		return;

	/* map two byte scancodes into one byte id's */

	break_flag = scancode > 0x7f;
	scancode &= 0x7f;
	if (scancode >= E0_BASE) {
#if 1
		printf("keyboard: scancode (%02x) not in range 00 - %2x\n", scancode, E0_BASE - 1);
#endif
		return;
	}

	if (kbd_dead(KGD_E0)) {
		int i;
		for (i = 0; i < sizeof(e0_keys); i++)
			if (scancode == e0_keys[i]) {
				scancode = E0_BASE + i;
				i = -1;
				break;
			}
		if (i != -1) {
#if 1
			printf("keyboard: unknown scancode e0 %02x\n", scancode);
#endif
			return;
		}
	}

	key_down[scancode].s = !break_flag;

	printf("keycode %3d %s\n",
		scancode,
		break_flag ? "release" : "press");
}
#endif

int fd;
struct termios old;

void clean_up(void) {
	if (ioctl(fd, KDSKBMODE, K_XLATE)) {
		perror("KDSKBMODE");
		exit(1);
	}
	tmp = tcsetattr(fd, 0, &old);
	if (tmp)
		printf("tcsetattr = %d\n", tmp);
	close(fd);
}

void die(int x) {
#if 0
	printf("caught signal %d, cleaning up...\n", x);
#endif
	clean_up();
	exit(1);
}

void watch_dog(int x) {
	die(x);
}

void usage(void) {
	fprintf(stderr, "\
showkey version " VERSION " (" OPTION ")

usage: showkey [options...]

valid options are:

	-h --help	display this help text
	-s --scancodes	display only the raw scan-codes
	-k --keycodes	display only the interpreted keycodes (default)
");
	exit(1);
}

main (int argc, char *argv[]) {
	const char *short_opts = "hsk";
	const struct option long_opts[] = {
		{ "help",	no_argument, NULL, 'h' },
		{ "scancodes",	no_argument, NULL, 's' },
		{ "keycodes",	no_argument, NULL, 'k' },
		{ NULL, 0, NULL, 0 }
	};
	int c;
	int show_keycodes = 1;

	struct termios new;
	unsigned char buf[16];
	int i, n;

	while ((c = getopt_long(argc, argv,
		short_opts, long_opts, NULL)) != -1) {
		switch (c) {
			case 's':
				show_keycodes = 0;
				break;
			case 'k':
				show_keycodes = 1;
				break;
			case 'h':
			case '?':
				usage();
		}
	}

	if (optind < argc)
		usage();

	/* the program terminates when there is no input for 10 secs */
	signal(SIGALRM, watch_dog);

	/*
		if we receive a signal, we want to exit nicely, in
		order not to leave the keyboard in an unusable mode
	*/
	signal(SIGHUP, die);
	signal(SIGINT, die);
	signal(SIGQUIT, die);
	signal(SIGILL, die);
	signal(SIGTRAP, die);
	signal(SIGABRT, die);
	signal(SIGIOT, die);
	signal(SIGFPE, die);
	signal(SIGKILL, die);
	signal(SIGUSR1, die);
	signal(SIGSEGV, die);
	signal(SIGUSR2, die);
	signal(SIGPIPE, die);
	signal(SIGTERM, die);
	signal(SIGSTKFLT, die);
	signal(SIGCHLD, die);
	signal(SIGCONT, die);
	signal(SIGSTOP, die);
	signal(SIGTSTP, die);
	signal(SIGTTIN, die);
	signal(SIGTTOU, die);

	if ((fd = open("/dev/console", O_RDONLY)) < 0) {
		perror("/dev/console");
		exit(1);
	}
	tmp = tcgetattr(fd, &old);
	if (tmp)
		printf("tcgetattr = %d\n", tmp);
	tmp = tcgetattr(fd, &new);
	if (tmp)
		printf("tcgetattr = %d\n", tmp);

	new.c_lflag = new.c_lflag & ~ (ICANON | ECHO | ISIG);
	new.c_cc[VMIN] = sizeof(buf);
	new.c_cc[VTIME] = 1;	/* 0.1 sec intercharacter timeout */

	tmp = tcsetattr(fd, TCSAFLUSH, &new);
	if (tmp)
		printf("tcsetattr = %d\n", tmp);
	if (ioctl(fd, KDSKBMODE,
#ifndef OLDKERNEL
		show_keycodes ? K_MEDIUMRAW : K_RAW)) {
#else
		K_RAW)) {
#endif
		perror("KDSKBMODE");
		exit(1);
	}

	printf("press any key (program terminates after 10s of last keypress)...\n");
	while (1) {
		alarm(10);
		n = read(fd, buf, sizeof(buf));
		for (i = 0; i < n; i++) {
			if (!show_keycodes)
				printf("0x%02x ", buf[i]);
#ifndef OLDKERNEL
			else
				printf("keycode %3d %s\n",
					buf[i] & 0x7f,
					buf[i] & 0x80 ? "release" : "press");
#else
			else {
				kbd_prev_dead_keys |= kbd_dead_keys;
				if (!kbd_dead_keys)
					kbd_prev_dead_keys = 0;
				kbd_dead_keys = 0;
				translate(buf[i]);
			}
#endif
		}
		if (!show_keycodes)
			printf("\n");
	}

	clean_up();
	exit(0);
}
